#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion
using System;
using System.Reflection;
using System.Text;

namespace HeuristicLab.Persistence.Default.CompositeSerializers.Storable.Descriptors {

  /// <summary>
  /// Describes the field of a type.
  /// </summary>
  [StorableClass]
  public sealed class FieldDescriptor : ITypeComponent, ICloneable {

    #region Fields & Properties

    /// <summary>
    /// The type that declared this field.
    /// </summary>
    [Storable]
    public TypeDescriptor DeclaringType { get; set; }

    /// <summary>
    /// The field name in the object graph as obtained by reflection.
    /// </summary>
    [Storable]
    public string RealName { get; set; }

    /// <summary>
    /// The name used for storing this field if different from the
    /// the real name.
    /// </summary>
    [Storable]
    public string StoredName { get; set; }

    /// <summary>
    /// The default value of this field (if any) when no value
    /// was found during deserialization.
    /// </summary>
    [Storable]
    public object DefaultValue { get; set; }

    /// <summary>
    /// The name used for (de)serialization which is either the
    /// StoredName (if present) or otherwise the RealName.
    /// </summary>
    public string Name { get { return StoredName ?? RealName; } }

    private FieldInfo handle;
    /// <summary>
    /// The FieldInfo as obtained by reflection representing the
    /// described type.
    /// </summary>
    public FieldInfo Handle {
      get {
        if (handle == null)
          handle = DeclaringType.Handle.GetField(RealName, BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic);
        return handle;
      }
    }
    #endregion

    [StorableConstructor]
    private FieldDescriptor(bool deserializing) { }

    /// <summary>
    /// Create a new field descriptor.
    /// </summary>
    /// <param name="declaringType">The type declaring this field.</param>
    /// <param name="fieldInfo">The field info obtained by reflection.</param>
    /// <param name="attribute">The storable attribute (if any) of this field.</param>
    public FieldDescriptor(TypeDescriptor declaringType, FieldInfo fieldInfo, StorableAttribute attribute) {
      DeclaringType = declaringType;
      RealName = fieldInfo.Name;
      if (attribute != null) {
        StoredName = attribute.Name;
        DefaultValue = attribute.DefaultValue;
      }
      handle = fieldInfo;
    }

    /// <summary>
    /// Create a string representation of this field descriptor.
    /// </summary>
    public override string ToString() {
      StringBuilder sb = new StringBuilder();
      sb.Append(RealName);
      if (StoredName != null)
        sb.AppendFormat(" ({0})", StoredName);
      if (DefaultValue != null)
        sb.AppendFormat(" [{0}]", DefaultValue);
      return sb.ToString();
    }

    /// <summary>
    /// Creates a shallow copy of this field descriptor.
    /// </summary>
    public object Clone() {
      var clone = new FieldDescriptor(false);
      clone.DeclaringType = DeclaringType;
      clone.RealName = RealName;
      clone.StoredName = StoredName;
      clone.DefaultValue = DefaultValue;
      clone.handle = handle;
      return clone;
    }

    /// <summary>
    /// From the given object return the value of the
    /// field described by this field descriptor.
    /// </summary>
    public object GetValue(object o) {
      return Handle.GetValue(o);
    }

    /// <summary>
    /// From the given object set the value of the
    /// field described by this field descriptor.
    /// </summary>
    public void SetValue(object o, object value) {
      Handle.SetValue(o, value);
    }

  }

}
